import unittest
from ansible_please.playbook import Playbook
from ansible_please.task_templates import Command
from ansible_please.ansible_dict import AnsibleDict
from .test_constants import EXPECTED_TEST_PLAYBOOK  # noqa


class TestPlaybook(unittest.TestCase):
    def test_playbook_creation(self):
        c = Command(cmd="echo test")
        p = Playbook(name="test_playbook", hosts="test_hosts", tasks=[c])
        assert p.asdict() == AnsibleDict(EXPECTED_TEST_PLAYBOOK)



from __future__ import annotations

import typing
import attr
from ansible_please.inventory import Inventory
from ansible_please.playbook import Playbook
from loguru import logger
import ansible_runner

# from ansible_runner.runner_config import RunnerConfig
from contextlib import contextmanager
import tempfile


class TempfileTuple:
    def __init__(self, inventory, playbook, delete=False):
        self.inventory_tempfile: tempfile.NamedTemporaryFile = inventory.save_to_file(
            delete
        )
        self.playbook_tempfile: tempfile.NamedTemporaryFile = playbook.save_to_file(
            delete
        )

    def delete(self) -> None:
        if self.delete:
            self.inventory_tempfile.close()
            self.playbook_tempfile.close()


@attr.s(repr=False)
class AnsibleRunner:

    playbooks: typing.List[Playbook] = attr.ib(default=[])
    input_path: str = attr.ib(default=None)
    inventory: Inventory = attr.ib(default=None)

    def __attrs_post_init__(self):
        if self.input_path is None and self.inventory is None:
            raise ValueError("Input path or Inventory must be provided!")
        if self.input_path is not None:
            self.inventory = Inventory.from_yaml(self.input_path)

    @contextmanager
    def _load_yamls(self, playbook: Playbook, delete_files=False) -> TempfileTuple:
        t = TempfileTuple(self.inventory, playbook, delete=delete_files)
        try:
            yield t
        finally:
            t.delete()

    def run(
        self, tags: typing.List[str] = ["up"], delete_files=True, verbosity=1
    ) -> None:
        tags = " ".join(tags)
        for playbook in self.playbooks:
            with self._load_yamls(playbook, delete_files=delete_files) as t:
                out = ansible_runner.run(
                    private_data_dir="/tmp/",
                    playbook=t.playbook_tempfile.name,
                    inventory=t.inventory_tempfile.name,
                    tags=tags,
                    verbosity=verbosity,
                )
                logger.info("{}: {}".format(out.status, out.rc))
                logger.info("Final status:")
                logger.info(out.stats)

    def up(self, delete_files=True):
        self.run(tags=["up"], delete_files=True)

    def down(self, delete_files=True):
        self.run(tags=["down"], delete_files=True)



import typing

from ansible_please.ansible_task import BaseAnsible


class Playbook(BaseAnsible):
    def __init__(
        self,
        name: str = None,
        hosts: str = None,
        gather_facts: bool = True,
        tasks: typing.List[BaseAnsible] = None,
        input_dict: typing.Dict[str, typing.Any] = {},
    ):
        super(Playbook, self).__init__(input_dict)
        self.name = name
        self.hosts = hosts
        self.gather_facts = gather_facts
        self.tasks = tasks



import unittest

from ansible_please.task_templates import DockerContainer, Pip
from ansible_please.ansible_dict import AnsibleDict
from omegaconf import OmegaConf
import pytest

EXPECTED_DOCKER_DOWN_DICT = AnsibleDict(
    dict(
        OmegaConf.create(
            """
- name: '[down] test-container'
  docker_container:
    name: test_container
    state: absent
    user: nobody
    keep_volumes: false
    detach: true
    tty: false
    interactive: false
    network_mode: host
    container_default_behavior: compatibility
  tags:
  - down
"""
        )[0]
    )
)


EXPECTED_PIP_DICT = AnsibleDict(
    dict(
        OmegaConf.create(
            """
- name: '[up] Install package via pip'
  command:
    cmd: "{{ ansible_python_interpreter }} -m pip install test-package"
  tags:
  - up
"""
        )[0]
    )
)


class TestDockerContainer(unittest.TestCase):
    def test_down_docker_container(self):
        d = DockerContainer(
            task_description="test-container",
            name="test_container",
            image="test-image",
            ports=[],
        ).down()
        assert d.asdict()

    def test_invalid_ports(self):
        with pytest.raises(ValueError):
            _ = DockerContainer(
                task_description="test-container",
                name="test_container",
                image="test-image",
                ports=[2, 5],
            )


class TestPip(unittest.TestCase):
    def test_pip_creation(self):
        p = Pip(packages=["test-package"])
        assert p.asdict()



import attr

from ansible_please.ansible_task import AnsibleTask


@attr.s(repr=False)
class Command(AnsibleTask):

    task_name: str = attr.ib(default="command", init=False)

    task_description = attr.ib(default="Install something with command line")
    cmd: str = attr.ib(default="")



from __future__ import annotations

import attr
import typing
from typing import Optional

from ansible_please.ansible_task import AnsibleTask


@attr.s(repr=False)
class DockerContainer(AnsibleTask):

    task_name: str = attr.ib(default="community.docker.docker_container", init=False)

    name: Optional[str] = attr.ib(
        default="", validator=attr.validators.instance_of(str)
    )
    image: Optional[str] = attr.ib(
        default=None,
        converter=attr.converters.optional(str),
        validator=attr.validators.optional(attr.validators.instance_of(str)),
    )
    command: Optional[str] = attr.ib(
        default=None,
        converter=attr.converters.optional(str),
        validator=attr.validators.optional(attr.validators.instance_of(str)),
    )
    ports: Optional[typing.List[str]] = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(list)),
    )
    state: Optional[str] = attr.ib(
        default=None,
        converter=attr.converters.optional(str),
        validator=attr.validators.optional(attr.validators.instance_of(str)),
    )
    user: Optional[str] = attr.ib(
        default="nobody",
        converter=attr.converters.optional(str),
        validator=attr.validators.optional(attr.validators.instance_of(str)),
    )
    keep_volumes: bool = attr.ib(
        default=False,
        converter=attr.converters.optional(bool),
        validator=attr.validators.optional(attr.validators.instance_of(bool)),
    )
    detach: bool = attr.ib(default=True, validator=attr.validators.instance_of(bool))
    tty: bool = attr.ib(default=False, validator=attr.validators.instance_of(bool))
    interactive: bool = attr.ib(
        default=False, validator=attr.validators.instance_of(bool)
    )
    network_mode: str = attr.ib(
        default="host", validator=attr.validators.instance_of(str)
    )
    env: Optional[typing.Dict[typing.Any, typing.Any]] = attr.ib(default=None)
    entrypoint: Optional[typing.List[str]] = attr.ib(default=None)
    container_default_behavior: Optional[str] = attr.ib(default="compatibility")

    @ports.validator
    def ports_validator(self, attribute, value) -> None:
        if isinstance(self.ports, list):
            if len(self.ports) == 0:
                self.ports = None
                return

            for v in self.ports:
                if not isinstance(v, str):
                    raise ValueError(
                        "expected string but got {} of type {}".format(v, type(v))
                    )

    def down(self) -> DockerContainer:
        return DockerContainer(
            task_description=self.task_description,
            name=self.name,
            user=self.user,
            keep_volumes=self.keep_volumes,
            state="absent",
        )



from ansible_please.ansible_runner import AnsibleRunner
from ansible_please.playbook import Playbook
from ansible_please.task_templates import Pip, DockerContainer

import redis
import attr
import docker
from docker.models.containers import Container
from urllib.error import HTTPError
from loguru import logger
import socket
import typing
import cloudpickle


def get_sockname(s: socket.socket):
    s.connect(("10.255.255.255", 1))
    return s.getsockname()[0]


def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        IP = get_sockname(s)
    except Exception as e:
        logger.exception("Caught exception while trying to get ip: {}".format(e))
        IP = "127.0.0.1"
    finally:
        s.close()
    return IP


@attr.s(repr=False)
class DockerCluster:
    """
        Manages the ansible docker cluster
    """

    ansible_runner: AnsibleRunner = attr.ib(default=None)
    force_teardown: True = attr.ib(default=True)

    _cluster_info_redis_runner: AnsibleRunner = attr.ib(init=False)
    _ip: str = attr.ib(init=False, default=get_ip())
    _redis_client: redis.Redis = attr.ib(init=False, default=None)

    def __attrs_post_init__(self):
        self._redis_client = redis.Redis(host=self._ip, port="6380")
        self._set_ansible_runner_from_cluster()

    @property
    def _playbook_cluster_info(self):
        pip_task = Pip(packages=["docker"])
        cluster_info_task = DockerContainer(
            task_description="start-cluster-info",
            name="ansible_please-cluster-info",
            image="redis:latest",
            command=" --port 6380",
            env={"REDIS_PORT": "6380"},
        )
        return Playbook(
            name="set up ansible_please-cluster-info",
            hosts=self._ip,
            tasks=[pip_task.up(), cluster_info_task.up(), cluster_info_task.down()],
        )

    def _set_ansible_runner_from_cluster(self):
        _ansible_runner = self._get_ansible_runner_from_cluster()
        if self.ansible_runner is not None:
            if _ansible_runner is not None:
                if self.force_teardown:
                    logger.info("Tearing down existing cluster...")
                    _ansible_runner.down(verbosity=0)
                    _cluster_info_redis_runner = AnsibleRunner(
                        playbooks=[self._playbook_cluster_info],
                        inventory=_ansible_runner.inventory,
                    )
                    _cluster_info_redis_runner.down(verbosity=0)
                    logger.info("Existing cluster torn down!")
                else:
                    logger.warn(
                        "Existing ansible_please cluster detected but force_teardown is set as False, this could lead to issues with untracked containers"
                    )
        else:
            self.ansible_runner = _ansible_runner
        if self.ansible_runner is None:
            raise ValueError("AnsibleRunner must not be None!")
        self._cluster_info_redis_runner = AnsibleRunner(
            playbooks=[self._playbook_cluster_info],
            inventory=self.ansible_runner.inventory,
        )

    def _get_ansible_runner_from_redis(self) -> AnsibleRunner:
        return cloudpickle.loads(self._redis_client.get("ansible_please_runner"))

    def _get_ansible_runner_from_cluster(self) -> AnsibleRunner:
        try:
            client: docker.DockerClient = docker.DockerClient()
            container: Container = client.containers.get("ansible_please-cluster-info")
            if container.attrs["State"]["Running"]:
                logger.info(
                    "Found existing cluster with container id {}".format(
                        container.attrs["Id"]
                    )
                )
                return self._get_ansible_runner_from_redis()
            return None
        except HTTPError as e:
            # Need to check its an 403 meaning that the container doesn't exist
            status_code = e.response.status_code
            if status_code == 404:
                return None
        except Exception as e:
            logger.debug(
                "Caught exception when attempting to get current state: {}. This may be expected".format(
                    e
                )
            )
            return None

    def run(
        self, tags: typing.List[str] = ["up"], delete_files=True, verbosity=1
    ) -> None:
        self.ansible_runner.run(
            tags=tags, delete_files=delete_files, verbosity=verbosity
        )

    def up(self, delete_files=True, verbosity=1):
        self._setup_cluster_info()
        self.run(tags=["up"], delete_files=delete_files, verbosity=verbosity)

    def down(self, delete_files=True, verbosity=1):
        self._cluster_info_redis_runner.run(
            tags=["down"], delete_files=True, verbosity=0
        )
        self.run(tags=["down"], delete_files=delete_files, verbosity=verbosity)

    def _setup_cluster_info(self):
        self._cluster_info_redis_runner.run(tags=["up"], delete_files=True, verbosity=0)
        self._redis_client.set(
            "ansible_please_runner", cloudpickle.dumps(self.ansible_runner)
        )



#!/usr/bin/env python

"""The setup script."""
import os
from setuptools import setup, find_packages

with open('requirements.txt') as f:
    install_reqs = f.read().splitlines()

for req in install_reqs:
    os.system("python -m pip install {}".format(req))
    
with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.md') as history_file:
    history = history_file.read()

requirements = ['Click>=7.0', 'ansible-runner', 'ansible', 'ansible-core==2.11.0', 'attrs', 'loguru', 'omegaconf==2.1.0.rc1', 'cloudpickle', 'docker']

setup_requirements = [ ]

test_requirements = ['docker', 'pytest']

setup(
    author="Shyam Sudhakaran",
    author_email='shyamsn97@gmail.com',
    python_requires='>=3.7',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Helper package to make running Ansible a bit smoother",
    long_description=readme + '\n\n' + history,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    license="MIT license",
    include_package_data=True,
    keywords='ansible_please',
    name='ansible_please',
    packages=find_packages(include=['ansible_please', 'ansible_please.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/shyamsn97/ansible_please',
    version='0.1.18',
    zip_safe=False,
)



from __future__ import annotations

import attr
from ansible_please.base_ansible import BaseAnsible
import typing
from ansible_please.ansible_dict import AnsibleDict
import types
import copy


def up(f, *args, **kwargs):
    def wrapped(*args, **kwargs) -> AnsibleTask:
        ansible_task_output = copy.deepcopy(f(*args, **kwargs))
        assert isinstance(ansible_task_output, AnsibleTask)
        ansible_task_output.task_state = "up"
        return ansible_task_output

    return wrapped


def down(f, *args, **kwargs):
    def wrapped(*args, **kwargs) -> AnsibleTask:
        ansible_task_output = copy.deepcopy(f(*args, **kwargs))
        assert isinstance(ansible_task_output, AnsibleTask)
        ansible_task_output.task_state = "down"
        return ansible_task_output

    return wrapped


class TaskMeta(type):
    def __new__(cls, name, bases, attr):
        task_states_map = {"up": up, "down": down}
        for name, value in attr.items():
            if name in task_states_map:
                if isinstance(value, types.FunctionType) or isinstance(
                    value, types.MethodType
                ):
                    attr[name] = task_states_map[name](value)

        return super(TaskMeta, cls).__new__(cls, name, bases, attr)


def validate_name(instance, attribute, value) -> None:
    if value == "":
        raise ValueError("Expected name to be a string of length > 0")


def validate_tags(instance, attribute, value) -> None:
    exclude_patterns = ["down", "up"]
    value = set(value)
    for pattern in exclude_patterns:
        value.discard(pattern)
    value = list(value)


@attr.s(repr=False)
class AnsibleTask(BaseAnsible, metaclass=TaskMeta):
    task_name: str = attr.ib(
        default="", validator=[attr.validators.instance_of(str), validate_name]
    )
    task_description: str = attr.ib(default="")
    task_state: str = attr.ib(default="up")
    tags: typing.List[str] = attr.ib(default=[], validator=[validate_tags])

    _exclude_patterns: typing.List[str] = attr.ib(
        init=False,
        default=[
            "task_name",
            "task_description",
            "tags",
            "task_state",
            "_exclude_patterns",
        ],
    )

    def asdict(self, additional_tags=[]) -> AnsibleDict:
        super_dict = super().asdict().data
        for pattern in self._exclude_patterns:
            super_dict.pop(pattern, None)
        return AnsibleDict(
            {
                "name": "[{}] {}".format(self.task_state, self.task_description),
                self.task_name: super_dict,
                "tags": self.tags + [self.task_state],
            }
        )

    def up(self) -> AnsibleTask:
        return self

    def down(self) -> AnsibleTask:
        return self

    def __str__(self):
        return self.asdict().__str__()



import typing
from typing import NamedTuple
from omegaconf import OmegaConf
from ansible_please.utils import save_yaml_to_file


class AnsibleDict(NamedTuple):
    data: typing.Dict[str, typing.Any]

    def asdict(self):
        return self._asdict()["data"]

    def pretty(self, use_list=True):
        d = self.asdict()
        if use_list:
            d = [d]
        return OmegaConf.to_yaml(OmegaConf.create(d))

    def save_to_file(self) -> str:
        """saves yaml to tempfile

        Returns:
            str: returns tempfile name
        """
        return save_yaml_to_file(self.__str__())

    def __str__(self):
        return self.pretty()
